open Datapoint
open Util
open SuffixTree
open SuffixTreeKernel

let reverse_vocab = ref (Vocabulary.reverse Vocabulary.global)

let nextValue x =
  if x < 8 then 8
  else int_of_float (2.0 ** ceil(log(float_of_int x) /. log(2.0)))

let output_float_array c a =
  output_binary_int c (Array.length a);
  for i=0 to Array.length a-1 do
    output_value c a.(i);
  done

let input_float_array c =
  let l = input_binary_int c in
  let a = Array.make l 0. in
    for i=0 to l-1 do
      a.(i) <- input_value c;
    done;
    a

let output_int_array c a =
  output_binary_int c (Array.length a);
  for i=0 to Array.length a-1 do
    output_binary_int c a.(i);
  done

let input_int_array c =
  let l = input_binary_int c in
  let a = Array.make l 0 in
    for i=0 to l-1 do
      a.(i) <- input_binary_int c;
    done;
    a

let output_binary_kernel c = function
    LinearKernel             -> 
      (output_byte c 0 ; output_value c 0 ; output_value c 0 ; output_value c 0)
  | PolynomialKernel (a,b,d) -> 
      (output_byte c 1 ; output_value c a ; output_value c b ; output_value c d)
  | RbfKernel g -> 
      (output_byte c 2 ; output_value c g ; output_value c 0 ; output_value c 0)
  | SigKernel (a,b) -> 
      (output_byte c 3 ; output_value c a ; output_value c b ; output_value c 0)
  | IdKernel b ->
      (output_byte c 4 ; output_value c b ; output_value c 0 ; output_value c 0)

let output_suffix_tree c t =
  output_binary_int c t.e;
  output_int_array  c t.tree_string;
  output_binary_int c t.num_nodes;

  let put = Array.make t.num_nodes false in
  let rec putNd nd =
    if put.(nd.identifier) then () else begin
      output_binary_int c nd.identifier;
      output_binary_int c nd.num_leaves;
      output_binary_int c nd.string_length;
      output_binary_int c nd.edge_label_start;
      output_binary_int c nd.edge_label_end;
      output_binary_int c nd.path_position;

      if isNull nd.sons 
      then output_binary_int c (-1) 
      else output_binary_int c (deref nd.sons).identifier;

      if isNull nd.father
      then output_binary_int c (-1) 
      else output_binary_int c (deref nd.father).identifier;

      if isNull nd.right_sibling 
      then output_binary_int c (-1) 
      else output_binary_int c (deref nd.right_sibling).identifier;

      if isNull nd.left_sibling 
      then output_binary_int c (-1) 
      else output_binary_int c (deref nd.left_sibling).identifier;
      
      if isNull nd.suffix_link
      then output_binary_int c (-1) 
      else output_binary_int c (deref nd.suffix_link).identifier;
      
      if not (isNull nd.sons) then putNd (deref nd.sons);
      if not (isNull nd.right_sibling) then putNd (deref nd.right_sibling);
    end in

    putNd (deref t.root);
    ()

let input_suffix_tree c = 
  let eval = input_binary_int c in
  let str  = input_int_array  c in
  let nn   = input_binary_int c in
  let nodes= Array.make nn nullP in
  let sonA = Array.make nn (-1) in
  let faA  = Array.make nn (-1) in
  let rA   = Array.make nn (-1) in
  let lA   = Array.make nn (-1) in
  let slA  = Array.make nn (-1) in
    for i=0 to nn-1 do
      let idn = input_binary_int c in
      let nl  = input_binary_int c in
      let strl= input_binary_int c in
      let els = input_binary_int c in
      let ele = input_binary_int c in
      let pp  = input_binary_int c in
        sonA.(idn) <- input_binary_int c;
        faA.(idn)  <- input_binary_int c;
        rA.(idn)   <- input_binary_int c;
        lA.(idn)   <- input_binary_int c;
        slA.(idn)  <- input_binary_int c;
        nodes.(idn) <- 
           newPtr { identifier = idn ; num_leaves = nl ; string_length = strl ;
                    edge_label_start = els ; edge_label_end = ele ; path_position = pp ;
                    sons = nullP ; right_sibling = nullP ; left_sibling = nullP ;
                    father = nullP ; suffix_link = nullP };
    done;
    for i=0 to nn-1 do
      if sonA.(i) != -1 then (deref nodes.(i)).sons <- nodes.(sonA.(i));
      if faA.(i)  != -1 then (deref nodes.(i)).father <- nodes.(faA.(i));
      if rA.(i)   != -1 then (deref nodes.(i)).right_sibling <- nodes.(rA.(i));
      if lA.(i)   != -1 then (deref nodes.(i)).left_sibling <- nodes.(lA.(i));
      if slA.(i)  != -1 then (deref nodes.(i)).suffix_link <- nodes.(slA.(i));
    done;
    { e = eval ; tree_string = str ; root = nodes.(0) ; num_nodes = nn }

(*
let output_suffix_tree c t =
  let rec findMaxId t =
    let m = ref t.identifier in
      for i=0 to Array.length t.sons-1 do
        m := max (!m) (findMaxId !(t.sons.(i)));
      done;
      !m in
  let maxId = findMaxId !t in
  let rec output_all nd =
    output_binary_int c nd.identifier;
    output_binary_int c nd.start_pos;
    output_binary_int c nd.end_pos;
    output_binary_int c nd.num_leaves;
    output_binary_int c nd.string_length;
    (match nd.slink with
	 None -> output_byte c 0
       | Some tr -> (output_byte c 1 ; output_binary_int c (!tr).identifier));

    output_binary_int c (Array.length nd.sons);
    for i=0 to Array.length nd.sons-1 do
      output_all !(nd.sons.(i));
    done in

    output_binary_int c maxId;
    output_all (!t);
    ()

let input_suffix_tree c = 
  let maxId = input_binary_int c in
  let arr   = Array.make (maxId+1) None in
  let slinks= Array.make (maxId+1) None in
  let rec input_all () =
    let id = input_binary_int c in
    let sp = input_binary_int c in
    let ep = input_binary_int c in
    let nl = input_binary_int c in
    let sl = input_binary_int c in
    let lnk= match input_byte c with 0 -> None | 1 -> Some (input_binary_int c) | _ -> internal_error "uncompatible suffix tree" in
    let k  = input_binary_int c in
    let nd0 = { identifier = -1; start_pos = -1; end_pos = -1; num_leaves = -1; string_length = -1; sons = [||]; father = None; slink = None } in
    let son= Array.init k (fun _ -> ref nd0) in
      for i = 0 to k - 1 do
	son.(i) <- input_all ();
      done;
      slinks.(id) <- lnk;
    let tr = ref
      { identifier=id; start_pos=sp; end_pos=ep; num_leaves=nl;
	string_length=sl; sons=son; father=None; slink=None} in
      arr.(id) <- Some tr;
      tr in
  let root = input_all () in
  let rec add_slinks_and_par p nd =
    (!nd).father <- p;
    (match slinks.((!nd).identifier) with
	 None -> ()
       | Some k -> (!nd).slink <- arr.(k));
    for i=0 to Array.length (!nd).sons - 1 do
      add_slinks_and_par (Some nd) (!nd).sons.(i);
    done in
  let _ = add_slinks_and_par None root in
    root
*)

let output_binary_params c = function
    DVP (w,k) -> 
      begin 
	output_byte c 0;
	output_value c w;
	output_binary_kernel c k;
	output_binary_int c 0;
	output_value c (0.0);
	output_binary_int c 0;
      end
  | SVP (w,k,i) ->
      begin 
	output_byte c 1;
	output_value c w;
	output_binary_kernel c k;
	output_binary_int c i;
	output_value c (0.0);
	output_binary_int c 0;
      end
  | DPP (w,l,n) ->
      begin
	output_byte c 2;
	output_value c w;
	output_binary_kernel c LinearKernel;
	output_binary_int c 0;
	output_value c l;
	output_binary_int c n;
      end
  | STP (w,f) ->
      begin
	output_byte c 3;
	output_value c w;
	output_binary_kernel c LinearKernel;

        (match f with
             ExpDecay l     -> (output_binary_int c 0 ; output_value c l ; output_binary_int c 0)
           | ConstWeight    -> (output_binary_int c 1 ; output_value c 0 ; output_binary_int c 0)
           | BoundedRange n -> (output_binary_int c 2 ; output_value c 0 ; output_binary_int c n));
      end
  | DiscP (w,i,b) ->
      begin
	output_byte c 4;
	output_value c w;
	output_binary_kernel c LinearKernel;
	output_binary_int c i;
	output_value c b;
	output_binary_int c 0;
      end
  | MatP (w,_) ->
      begin
	output_byte c 5;
	output_value c w;
	output_binary_kernel c LinearKernel;
	output_binary_int c 0;
	output_value c (0.0);
	output_binary_int c 0;
      end
  | FVParamPlaceHolder ->
      begin
	output_byte c 6;
	output_value c (0.0);
	output_binary_kernel c LinearKernel;
	output_binary_int c 0;
	output_value c (0.0);
	output_binary_int c 0;
      end

let output_clump c = function
    DV v ->
      begin
	output_byte c 0;
	output_binary_int c (Array.length v);
	for i=0 to Array.length v - 1 do
	  output_value c v.(i);
	done;
      end
  | SV (f,v) -> 
      begin
	output_byte c 1;
	let cnt = ref 0 in
        let hitNeg = ref false in
	for i=0 to Array.length f - 1 do
          if f.(i) < 0 then hitNeg := true;
	  if nonZero v.(i) && not !hitNeg then cnt := !cnt + 1;
	done;
	output_binary_int c !cnt;
        hitNeg := false;
	for i=0 to Array.length f - 1 do
          if f.(i) < 0 then hitNeg := true;
	  if nonZero v.(i) && not !hitNeg then 
	    (output_binary_int c f.(i);
	     output_value      c v.(i););
	done;
      end
  | DP v -> 
      begin
	output_byte c 2;
	output_binary_int c (Array.length v);
	for i=0 to Array.length v - 1 do
	  output_value c (!reverse_vocab).(v.(i));
	done;
      end
  | ST st -> 
      begin
	output_byte c 3;
        output_suffix_tree c st.tree_ref;
        output_float_array c st.tree_kvc;
      end
  | Disc i -> 
      begin
	output_byte c 4;
	output_value c (!reverse_vocab).(i);
      end
  | Mat i ->
      begin
	output_byte c 5;
	output_binary_int c i;
      end


let saveDataSetTo c dataset = 
  let _ = reverse_vocab := Vocabulary.reverse Vocabulary.global in
    (* ranking information *)
    (match dataset.rankingPoints with
	None   -> output_byte c 0
      | Some a -> 
	  begin
	    output_byte c 1;
	    output_binary_int c (Array.length a);
	    for i=0 to Array.length a - 1 do
	      output_binary_int c a.(i);
	    done;
	  end);

    (* now, the training parameters; we want to make sure we have
       enough space to add more classes if we need, so, since this is
       relatively small (in comparison), we write up to the next
       power of two, with at least 8, i.e.,
         1-3 -> 8
         4-7 -> 8
         8-15 -> 16
         16-31 -> 32
         etc... *)
    let numClumps = Array.length dataset.trainingParams in
    output_binary_int c numClumps;
    for i = 0 to numClumps-1 do
      output_binary_params c dataset.trainingParams.(i);
    done;
    for i = numClumps to nextValue numClumps do
      output_binary_params c FVParamPlaceHolder;
    done;

    (* the above should take up (nearly) constant space; the remainder will not *)
    output_string c "Data\n";
    
    (* first, how many data points *)
    output_value c dataset.dataHasClasses;
    let ntp = Array.length dataset.datapoints in

    output_binary_int c ntp;
    output_byte c (if isNone dataset.datanormalize then 0 else 1);

    for i=0 to ntp-1 do
      (* output the class, if we have one *)
      if dataset.dataHasClasses then output_value c dataset.dataclasses.(i);
      (* output the data *)
      for j=0 to Array.length dataset.datapoints.(i)-1 do
	output_clump c dataset.datapoints.(i).(j);
      done;
    done;

    (* now, if we have any matrices, we need to dump those too *)
    for i = 0 to numClumps-1 do
      (match dataset.datanormalize with
           None -> ()
         | Some a -> output_float_array c a.(i));

      match dataset.trainingParams.(i) with
	  MatP (_,m) -> 
	    begin
	      output_binary_int c 999;
	      let l = (Array.length m) in
	      output_binary_int c l;
	      for i=0 to l-1 do
		for j=0 to i do
		  output_value c m.(i).(j);
		done;
	      done;
	    end
	| _ -> ();
    done;

    (* write x-val data *)
    (match dataset.crossValSplit with
	 None -> output_byte c 0
       | Some (a,n) -> (
	   output_byte c 1;
           output_binary_int c n;
	   output_binary_int c (Array.length a);
	   for i=0 to Array.length a-1 do
	     output_binary_int c a.(i);
	   done;));
    ()

let saveDataSet filepath dataset =
  let c = open_out_bin filepath in
    (* write the header *)
    output_string c "#SVMsequel data\n";
    saveDataSetTo c dataset;
    output_string c "End of file\n";
    close_out c


let input_binary_kernel ch = 
  let a = input_byte  ch in
  let b = input_value ch in
  let c = input_value ch in
  let d = input_value ch in
    match a with
	0 -> LinearKernel
      | 1 -> PolynomialKernel (b,c,d)
      | 2 -> RbfKernel b
      | 3 -> SigKernel (b,c)
      | 4 -> IdKernel b
      | _ -> internal_error "input_binary_kernel: malformed input"

let input_clump c = 
  match input_byte c with
      0 -> 
	begin
	  let l = input_binary_int c in
	  let v = Array.make l 0.0 in
	    for i=0 to l-1 do
	      v.(i) <- input_value c;
	    done;
	    DV v
	end
    | 1 -> 
	begin
	  let l = input_binary_int c in
	  let f = Array.make l 0 in
	  let v = Array.make l 0.0 in
	    for i=0 to l-1 do
	      f.(i) <- input_binary_int c;
	      v.(i) <- input_value c;
	    done;
	    SV (f,v)
	end
    | 2 ->
	begin
	  let l = input_binary_int c in
	  let v = Array.make l 0 in
	    for i=0 to l-1 do
	      let s = input_value c in
                v.(i) <- Vocabulary.get Vocabulary.global s;
	    done;
	    DP v
	end
    | 3 -> 
        begin
          let tree = input_suffix_tree c in
          let kvc  = input_float_array c in
            ST { tree_ref = tree ; tree_kvc = kvc }
        end
    | 4 -> Disc (Vocabulary.get Vocabulary.global (input_value c))
    | 5 -> Mat  (input_binary_int c)
    | n -> internal_error ("input_clump malformed input, got " ^ string_of_int n)

let input_binary_params ch = 
  let a = input_byte ch in
  let (b : float) = input_value ch in
  let c = input_binary_kernel ch in
  let d = input_binary_int ch in
  let (e : float) = input_value ch in
  let f = input_binary_int ch in
    match a with
	0 -> DVP (b, c)
      | 1 -> SVP (b, c, d)
      | 2 -> DPP (b, e, f)
      | 3 -> STP (b, match d with 0 -> ExpDecay e | 1 -> ConstWeight | 2 -> BoundedRange f | _ -> internal_error "unknown STP type")
      | 4 -> DiscP (b, d, e)
      | 5 -> MatP (b, [||])
      | 6 -> FVParamPlaceHolder
      | _ -> internal_error "input_binary_params: malformed input"

let loadDataSetFrom c =
  let isRanking = input_byte c == 1 in
  let rankingArray =
    if isRanking 
    then 
      begin
	let l = input_binary_int c in
	let a = Array.make l 0 in
	  for i=0 to l-1 do
	    a.(i) <- input_binary_int c;
	  done;
	  Some a
      end
    else None in
  let numClumps = input_binary_int c in
  let tps = Array.make numClumps FVParamPlaceHolder in
    for i = 0 to numClumps-1 do
      tps.(i) <- input_binary_params c;
    done;
    for i = numClumps to nextValue numClumps do
      let _ = input_binary_params c in ();
    done;

  let lne = input_line c in
    if lne <> "Data" then failwith ("Data file appears to be corrupted");
  let hasC = input_value c in
  let ntp  = input_binary_int c in
  let hasNorm = input_byte c == 1 in
  let norm = if hasNorm then Some (Array.init numClumps (fun _ -> [||])) else None in
  let cl = if hasC then Array.make ntp 0.0 else [| |] in
  let dp = Array.make ntp [||] in
    for i=0 to ntp-1 do
      if hasC then cl.(i) <- input_value c;
      dp.(i) <- Array.make numClumps (Mat (-1));
      for j=0 to numClumps-1 do
	dp.(i).(j) <- input_clump c;
      done;
    done;

  (* finally, read the matrices *)
    for i = 0 to numClumps-1 do
      (match norm with
           None -> ()
         | Some a -> a.(i) <- input_float_array c);

      match tps.(i) with
	  MatP (w,_) ->
	    begin
	      if input_binary_int c != 999 then failwith "Data file appears to be corrupted";
	      let l = input_binary_int c in
	      let m = Array.init l (function i -> Array.make (i+1) 0.0) in
		for i=0 to l-1 do
		  for j=0 to i do
		    m.(i).(j) <- input_value c;
		  done;
		done;
		tps.(i) <- MatP (w,m);
	    end
	| _ -> ();
    done;

    let cvs = 
      if input_byte c == 0 then None else begin
        let nfold = input_binary_int c in
	let l = input_binary_int c in
	let a = Array.make l 0 in
	  for i = 0 to l-1 do
	    a.(i) <- input_binary_int c;
	  done;
	  Some (a,nfold)
      end in

    { trainingParams = tps;
      datapoints = dp;
      dataclasses = cl;
      datanormalize = norm;
      rankingPoints = rankingArray;
      crossValSplit = cvs;
      dataHasClasses = hasC
    }

let loadDataSet filepath =
  try 
    let c = open_in_bin filepath in
      (* read the header *)
    let lne = input_line c in
      if lne <> "#SVMsequel data" then internal_error ("File does not appear to be SVMsequel data, wrong header, loading data failed");
      let ds = loadDataSetFrom c in
      let lne = input_line c in
        if lne <> "End of file" then internal_error ("Error at end of file, appears to be corrupted")
          close_in c;
        ds
  with 
      End_of_file -> internal_error ("Reached end of data file prematurely")
    | e -> raise e
    

(* loading and saving models follows *)
let saveModelTo c m =
  let _ = reverse_vocab := Vocabulary.reverse Vocabulary.global in
  output_binary_int c (Array.length m.modelParams);
  output_binary_int c (Array.length m.supportVectors);
  
  output_byte c (if isNone m.modelnormalize then 0 else 1);
  
  for i=0 to Array.length m.modelParams-1 do
    output_binary_params c m.modelParams.(i);
    (match m.modelnormalize with
         None -> ()
       | Some a -> output_float_array c a.(i));
  done;
  
  for i=0 to Array.length m.supportVectors-1 do
    for j=0 to Array.length m.modelParams-1 do
      output_clump c m.supportVectors.(i).(j);
    done;
  done;
  
  output_binary_int c (Array.length m.multipliers);
  for i = 0 to Array.length m.multipliers-1 do
    output_float_array c m.multipliers.(i);
  done;
  
  output_float_array c m.modelBias;
    
  (match m.multiclass with
       None -> output_byte c 0
     | Some (OVA numC) -> ( output_byte c 1 ; output_int_array c numC )
     | Some (OVN numC) -> ( output_byte c 2 ; output_int_array c numC )
     | Some (AVA numC) -> ( output_byte c 3 ; output_int_array c numC ));


  (match m.binarySigmoid with
       None -> output_byte c 0
     | Some (a,b) -> ( output_byte c 1 ; output_value c a ; output_value c b ));

  (match m.multiSigmoid with
       None -> output_byte c 0
     | Some a -> ( output_byte c 1 ; output_float_array c a ));

  ()

let saveModel filepath (m : modelData) =
  let c = open_out_bin filepath in
    output_string c "#SVMsequel model\n";
    saveModelTo c m;
    output_string c "End of model\n";
    close_out c;
    ()

let loadModelFrom c =
  let numClumps = input_binary_int c in
  let numSVS    = input_binary_int c in
  let hasNorm   = (input_byte c == 1) in
  let norm      = if hasNorm then Some (Array.init numClumps (fun _ -> [||])) else None in
  let p   = Array.make numClumps FVParamPlaceHolder in
  let svs = Array.init numSVS (function i -> Array.make numClumps (Mat (-1))) in
    for i=0 to numClumps-1 do
      p.(i) <- input_binary_params c;
      (match norm with
           None -> ()
         | Some a -> a.(i) <- input_float_array c);
    done;
    for i=0 to numSVS-1 do
      for j=0 to numClumps-1 do
        svs.(i).(j) <- input_clump c;
      done;
    done;

  let numMults = input_binary_int c in
  let m = Array.init numMults (function i -> [||]) in
    for i=0 to numMults-1 do
      m.(i) <- input_float_array c;
    done;

  let bias = input_float_array c in

  let mclass = 
    match input_byte c with
        0 -> None
      | 1 -> Some (OVA (input_int_array c))
      | 2 -> Some (OVN (input_int_array c))
      | 3 -> Some (AVA (input_int_array c))
      | _ -> failwith "model file error, invalid multiclass type" in


  let binSig =
    match input_byte c with
        0 -> None
      | 1 -> Some (input_value c, input_value c)
      | _ -> failwith "model file error, invalid binary sigmoid data" in

  let mulSig =
    match input_byte c with
        0 -> None
      | 1 -> Some (input_float_array c)
      | _ -> failwith "model file error, invalid multiclass sigmoid data" in
    
    { modelParams = p ;
      supportVectors = svs ;
      multipliers = m ;
      modelBias = bias;
      multiclass = mclass ;
      modelnormalize = norm;
      binarySigmoid = binSig ;
      multiSigmoid = mulSig
    }

let loadModel filepath =
  let c = open_in_bin filepath in
    if input_line c <> "#SVMsequel model" then failwith "malformed header in model file; file not loaded";
  let m = loadModelFrom c in
    if input_line c <> "End of model" then failwith "model file error";
    m

let saveEnvironment filepath dats mods vals =
  let c = open_out_bin filepath in
    output_string c "#SVMsequel environment\n";
    output_binary_int c (hash_size dats);
    output_binary_int c (hash_size mods);
    output_binary_int c (hash_size vals);
    Hashtbl.iter (fun s d -> output_value c s ; saveDataSetTo c d) dats;
    Hashtbl.iter (fun s m -> output_value c s ; saveModelTo   c m) mods;
    Hashtbl.iter (fun s v -> output_value c s ; output_value  c v) vals;
    output_string c "End of environment\n";
    close_out c;
    ()

let loadEnvironment filepath dats mods vals =
  let c = open_in_bin filepath in
    if input_line c <> "#SVMsequel environment" then failwith "malformed header in environment file; not loaded";
    let numD = input_binary_int c in
    let numM = input_binary_int c in
    let numV = input_binary_int c in
      for i=1 to numD do
        let s = input_value c in
        let d = loadDataSetFrom c in
          Hashtbl.replace dats s d;
      done;
      for i=1 to numM do
        let s = input_value c in
        let m = loadModelFrom c in
          Hashtbl.replace mods s m;
      done;
      for i=1 to numV do
        let s = input_value c in
        let v = input_value c in
          Hashtbl.replace vals s v;
      done;
      close_in c;
      ()

