open Array
open Datapoint

let binary_classify
      (p  : fvParams array)
      (sv : fvClump array array)
      (m  : float array)
      (x  : fvClump array)
      : float = 0.0

(* the return value of inferOne is a tuple of the individual
   components of the inference (for AVA, this is x_ij for all i>j; for
   OVA, this is x_i for all i; for binary this is just a singleton
   array.  the second component of the tuple is the chosen class *)
let inferOne 
      (model : modelData)
      (x     : fvClump array)
      : float array * int =
  match model.multiclass with
      None -> (* this should be binary classification *)
	let v = binary_classify 
		  model.modelParams 
		  model.supportVectors 
		  model.multipliers.(0) 
		  x
	in  ([| v |], if v > model.modelBias then 1 else -1)

    | Some (OVA numC) -> (* one versus all *)
	begin
	  let a = make numC 0.0 in
	  let maxV = ref 0.0 in
	  let maxP = ref (-1) in
	  for i=0 to numC-1 do
	    a.(i) <- binary_classify
	               model.modelParams
	               model.supportVectors
	               model.multipliers.(i)
	               x;
	    if i == 0 || a.(i) >= !maxV then ( maxV := a.(i); maxP := i; )
	  done;
	  (a, !maxP)
	end

    | Some (AVA numC) -> (* all versus all *)
	begin
	  let a = make (numC * (numC-1) / 2) 0.0 in
	  let votes = make numC 0 in
	  let maxV = ref 0 in
	  let maxP = ref (-1) in
	  for i=0 to numC-1 do
	    for j=i+1 to numC-1 do
	      a.(i*numC+j) 
	        <- binary_classify
	             model.modelParams
	             model.supportVectors
	             model.multipliers.(i*numC+j)
	             x;
	      let p = if a.(i*numC+j) > model.modelBias then i else j in
		votes.(p) <- votes.(p) + 1;
		if votes.(p) > !maxV then (maxV := votes.(p); maxP := p);
	    done;
	  done;
	    (a, !maxP)
	end
	
