open Array
open Datapoint
open Util

let binary_classify
      (p  : fvParams array)
      (sv : fvClump array array)
      (m  : float array)
      (x  : fvClump array)
      (norm: float array array option)
      : float =
  let v = ref 0.0 in
    for n = 0 to length sv-1 do
      if nonZero m.(n) then
	let d = Dot.dotClumps norm sv.(n) x n (-1) p in
	v := !v +. m.(n) *. d;
    done;
    !v

(* the return value of predictOne is a tuple of the individual
   components of the predictence (for AVA, this is x_ij for all i>j; for
   OVA, this is x_i for all i; for binary this is just a singleton
   array.  the second component of the tuple is the chosen class *)
let predictOne 
      (model : modelData)
      (x     : fvClump array)
      (norm: float array array option)
      : float array * int =
  match model.multiclass with
      None -> (* this should be binary classification *)
	let v = binary_classify 
		  model.modelParams 
		  model.supportVectors 
		  model.multipliers.(0) 
		  x
                  norm
	in  ([| v |], if v > model.modelBias.(0) then 1 else -1)

    | Some (OVA carr) 
    | Some (OVN carr) -> (* one versus all or one versus none *)
	begin
          let numC = length carr in
	  let a = make numC 0.0 in
	  let maxV = ref 0.0 in
	  let maxP = ref (-1) in
	  for i=0 to numC-1 do
	    a.(i) <- binary_classify
	               model.modelParams
	               model.supportVectors
	               model.multipliers.(i)
	               x
                       norm;
	    if i == 0 || a.(i) >= !maxV then ( maxV := a.(i); maxP := i; )
	  done;
	  (a, carr.(!maxP))
	end

    | Some (AVA carr) -> (* all versus all *)
	begin
          let numC = length carr in
	  let a = make (numC * (numC-1) / 2) 0.0 in
	  let votes = make numC 0 in
	  let maxV = ref 0 in
	  let maxP = ref (-1) in
	  for i=1 to numC-1 do
	    for j=0 to i-1 do 
	      a.(avaCLposition i j)
	        <- binary_classify
	             model.modelParams
	             model.supportVectors
	             model.multipliers.(avaCLposition i j)
	             x norm;
	      let p = if a.(avaCLposition i j) > 0. then i else j in
		votes.(p) <- votes.(p) + 1;
		if votes.(p) > !maxV then (maxV := votes.(p); maxP := p);
	    done;
	  done;
	    (a, carr.(!maxP))
	end

let predictCV
      (model : modelData)
      (d     : dataSet)
      (set   : int)      (* assume j!=set has been used for training *)
      : int =
  if not d.dataHasClasses then failwith "predictCV called on data without classes" else
    match d.crossValSplit with
        None -> failwith "predictCV called when cross validation is not turned on"
      | Some (crossVal,nfold) ->
          let err = ref 0 in
            for i=0 to length crossVal-1 do
              if crossVal.(i) == set then
                let (_,p) = predictOne model d.datapoints.(i) d.datanormalize in
                if p != int_of_float d.dataclasses.(i) then err := !err+1;
            done;
            !err

let predictAll
      (model : modelData)
      (d     : dataSet) =
  let err = ref 0 in
  let a   = init (length d.datapoints) (fun _ -> [||]) in
    for i=0 to length d.datapoints - 1 do
      Dot.clearCache ();
      let (arr,p) = predictOne model d.datapoints.(i) d.datanormalize in
        if p != int_of_float d.dataclasses.(i) then err := !err+1;
        a.(i) <- arr;
    done;
    (!err, a)

