open Datapoint
open Array
open Util

let trainBinarySigmoid (m : modelData) (d : dataSet) =
  if isSome m.multiclass then failwith "cannot train a binary sigmoid on multiclass data" else
  if not d.dataHasClasses then failwith "cannot train a sigmoid on data without classes" else
    let (_, arr) = Predict.predictAll m d in

    let label i  = if d.dataclasses.(i) > 0. then true else false in
    let deci  i  = arr.(i).(0) in

    let len      = length d.dataclasses in
    let (prior0, prior1) =
      let tmp0 = ref 0. in let tmp1 = ref 0. in
        for i=0 to len-1 do
          if label i then tmp1 := !tmp1 +. 1. else tmp0 := !tmp0 +. 1.;
        done;
        (!tmp0, !tmp1) in

    let maxiter = 104 in let minstep = 1e-10 in let sigma = 1e-3 in

    let hiTarget = (prior1 +. 1.) /. (prior1 +. 2.) in
    let loTarget = 1. /. (prior0 +. 2.) in

    let t i = if label i then hiTarget else loTarget in

    let curchar = ref 'a' in
    let next_char () = 
      match !curchar with
          'Z' -> 'Z'
        | 'z' -> curchar := 'A' ; 'z'
        | c   -> curchar := (Char.chr (1+Char.code c)) ; c in

    let vA = ref 0. in
    let vB = ref (log(prior0 +. 1.) /. (prior1 +. 1.)) in
    let fval = ref 0. in

      for i = 0 to len - 1 do
        let fApB = deci i *. !vA +. !vB in
        if fApB >= 0.
        then fval := !fval +. (t i      ) *. fApB +. log(1. +. exp(0. -. fApB))
        else fval := !fval +. (t i -. 1.) *. fApB +. log(1. +. exp(      fApB));
      done;

      (try
        for it = 1 to maxiter do
          let _   = if it mod 2 == 0 then ( print_char (next_char ()) ; flush stdout ) in
          let h11 = ref sigma in let h22 = ref sigma in
          let h21 = ref 0. in let g1 = ref 0. in let g2 = ref 0. in
            
            for i = 0 to len-1 do
              let fApB = deci i *. !vA +. !vB in
              let p =
                if fApB >= 0.
                then exp(0. -. fApB) /. (1. +. exp(0. -. fApB))
                else 1.              /. (1. +. exp(      fApB)) in
              let q =
                if fApB < 0.
                then exp(      fApB) /. (1. +. exp(      fApB))
                else 1.              /. (1. +. exp(0. -. fApB)) in
                
              let d2 = p *. q in
                h11 := !h11 +. deci i *. deci i *. d2;
                h22 := !h22 +. d2;
                h21 := !h21 +. deci i *. d2;
                g1  := !g1  +. deci i *. (t i -. p);
                g2  := !g2  +. t i -. p;
            done;
            
            if fabs !g1 < 1e-5 && fabs !g2 < 1e-5 then raise Break;

            let det = !h11 *. !h22 -. !h21 *. !h22 in
            let dA  = 0. -. (!h22 *. !g1 -. !h21 *. !g2) /. det in
            let dB  = 0. -. (!h11 *. !g2 -. !h21 *. !g1) /. det in
            let gd  = !g1 *. dA +. !g2 *. dB in
            let stepsize = ref 1. in
            let alldone = ref false in

              while not !alldone && (!stepsize >= minstep) do
                let newA = !vA +. !stepsize *. dA in
                let newB = !vB +. !stepsize *. dB in
                let newf = ref 0. in

                  for i = 0 to len - 1 do
                    let fApB = deci i *. newA +. newB in
                      if fApB >= 0.
                      then newf := !newf +. (t i      ) *. fApB +. log(1. +. exp(0. -. fApB))
                      else newf := !newf +. (t i -. 1.) *. fApB +. log(1. +. exp(      fApB))
                  done;

                  if !newf < !fval +. 0.0001 *. !stepsize *. gd
                  then ( vA := newA ; vB := newB ; fval := !newf ; alldone := true )
                  else ( stepsize := !stepsize /. 2. ;
                         if !stepsize < minstep then internal_error "training binary sigmoid parameters failed" )
              done;
        done
       with Break -> () | e -> raise e);

      (!vA, !vB)
